@@ -1,11 +1,14 @@
 # -*- coding: utf-8 -*-
 
+import fnmatch
 import ctypes
+import errno
 import glob
 import locale
 import os
 import re
 import shutil
+import struct
 import subprocess as sp
 import sys
 import tempfile
@@ -16,6 +19,9 @@ if sys.platform not in ("darwin", "win32"):
 	import grp
 	import pwd
 
+if sys.platform != "win32":
+	import fcntl
+
 try:
 	reloaded
 except NameError:
@@ -28,7 +34,8 @@ else:
 	warnings.warn("Module %s is being reloaded. This is NOT recommended." %
 				  __name__, RuntimeWarning)
 	warnings.warn("Implicitly reloading builtins", RuntimeWarning)
-	reload(__builtin__)
+	if sys.platform == "win32":
+		reload(__builtin__)
 	warnings.warn("Implicitly reloading os", RuntimeWarning)
 	reload(os)
 	warnings.warn("Implicitly reloading os.path", RuntimeWarning)
@@ -40,12 +47,18 @@ else:
 if sys.platform == "win32":
 	from win32file import *
 	from winioctlcon import FSCTL_GET_REPARSE_POINT
+	import win32file
+	import win32con
+	import pywintypes
+	import winerror
 
+# Cache used for safe_shell_filter() function
+_cache = {}
+_MAXCACHE = 100
 
 FILE_ATTRIBUTE_REPARSE_POINT = 1024
-SYMBOLIC_LINK = 'symbolic'
-MOUNTPOINT = 'mountpoint'
-GENERIC = 'generic'
+IO_REPARSE_TAG_MOUNT_POINT = 0xA0000003  # Junction
+IO_REPARSE_TAG_SYMLINK = 0xA000000C
 
 from encoding import get_encodings
 
@@ -211,13 +224,55 @@ def quote_args(args):
 	return args_out
 
 
-def dlopen(name):
+def dlopen(name, handle=None):
 	try:
-		return ctypes.CDLL(name)
+		return ctypes.CDLL(name, handle=handle)
 	except:
 		pass
 
 
+def find_library(pattern, arch=None):
+	"""
+	Use ldconfig cache to find installed library.
+	
+	Can use fnmatch-style pattern matching.
+	
+	"""
+	try:
+		p = sp.Popen(["/sbin/ldconfig", "-p"], stdout=sp.PIPE)
+		stdout, stderr = p.communicate()
+	except:
+		return
+	if not arch:
+		try:
+			p = sp.Popen(["file", "-L", sys.executable], stdout=sp.PIPE)
+			file_stdout, file_stderr = p.communicate()
+		except:
+			pass
+		else:
+			# /usr/bin/python2.7: ELF 64-bit LSB shared object, x86-64,
+			# version 1 (SYSV), dynamically linked, interpreter
+			# /lib64/ld-linux-x86-64.so.2, for GNU/Linux 3.2.0,
+			# BuildID[sha1]=41a1f0d4da3afee8f22d1947cc13a9f33f59f2b8, stripped
+			parts = file_stdout.split(",")
+			if len(parts) > 1:
+				arch = parts[1].strip()
+	for line in stdout.splitlines():
+		# libxyz.so (libc6,x86_64) => /lib64/libxyz.so.1
+		parts = line.split("=>", 1)
+		candidate = parts[0].split(None, 1)
+		if len(parts) < 2 or len(candidate) < 2:
+			continue
+		info = candidate[1].strip("( )").split(",")
+		if arch and len(info) > 1 and info[1].strip() != arch:
+			# Skip libs for wrong arch
+			continue
+		filename = candidate[0]
+		if fnmatch.fnmatch(filename, pattern):
+			path = parts[1].strip()
+			return path
+
+
 def expanduseru(path):
 	""" Unicode version of os.path.expanduser """
 	if sys.platform == "win32":
@@ -334,9 +389,9 @@ def get_program_file(name, foldername):
 	""" Get path to program file """
 	if sys.platform == "win32":
 		paths = getenvu("PATH", os.defpath).split(os.pathsep)
-		paths += glob.glob(os.path.join(getenvu("PROGRAMFILES", ""),
+		paths += safe_glob(os.path.join(getenvu("PROGRAMFILES", ""),
 										foldername))
-		paths += glob.glob(os.path.join(getenvu("PROGRAMW6432", ""),
+		paths += safe_glob(os.path.join(getenvu("PROGRAMW6432", ""),
 										foldername))
 		exe_ext = ".exe"
 	else:
@@ -425,7 +480,7 @@ def launch_file(filepath):
 		kwargs["startupinfo"].wShowWindow = sp.SW_HIDE
 		kwargs["shell"] = True
 		kwargs["close_fds"] = True
-		retcode = sp.call('start "" /B "%s"' % filepath, **kwargs)
+		retcode = sp.call('start "" "%s"' % filepath, **kwargs)
 	elif which('xdg-open'):
 		retcode = sp.call(['xdg-open', filepath], **kwargs)
 	return retcode
@@ -460,6 +515,39 @@ def mkstemp_bypath(path, dir=None, text=False):
 	return tempfile.mkstemp(ext, fname + "-", dir, text)
 
 
+def mksfile(filename):
+	"""
+	Create a file safely and return (fd, abspath)
+	
+	If filename already exists, add '(n)' as suffix before extension (will
+	try up to os.TMP_MAX or 10000 for n)
+	
+	Basically, this works in a similar way as _mkstemp_inner from the
+	standard library 'tempfile' module.
+	
+	"""
+
+	flags = tempfile._bin_openflags
+
+	fname, ext = os.path.splitext(filename)
+
+	for seq in xrange(tempfile.TMP_MAX):
+		if not seq:
+			pth = filename
+		else:
+			pth = "%s(%i)%s" % (fname, seq, ext)
+		try:
+			fd = os.open(pth, flags, 0600)
+			tempfile._set_cloexec(fd)
+			return (fd, os.path.abspath(pth))
+		except OSError, e:
+			if e.errno == errno.EEXIST:
+				continue  # Try again
+			raise
+
+	raise IOError, (errno.EEXIST, "No usable temporary file name found")
+
+
 def movefile(src, dst, overwrite=True):
 	""" Move a file to another location.
 	
@@ -484,7 +572,7 @@ def putenvu(name, value):
 		os.environ[name] = value.encode(fs_enc)
 
 
-def parse_reparse_buffer(original, reparse_type=SYMBOLIC_LINK):
+def parse_reparse_buffer(buf):
 	""" Implementing the below in Python:
 
 	typedef struct _REPARSE_DATA_BUFFER {
@@ -514,73 +602,34 @@ def parse_reparse_buffer(original, reparse_type=SYMBOLIC_LINK):
 	} REPARSE_DATA_BUFFER, *PREPARSE_DATA_BUFFER;
 
 	"""
-	# Size of our data types
-	SZULONG = 4 # sizeof(ULONG)
-	SZUSHORT = 2 # sizeof(USHORT)
+	# See https://docs.microsoft.com/en-us/windows-hardware/drivers/ddi/content/ntifs/ns-ntifs-_reparse_data_buffer
 
-	# Our structure.
-	# Probably a better way to iterate a dictionary in a particular order,
-	# but I was in a hurry, unfortunately, so I used pkeys.
-	buffer = {
-		'tag' : SZULONG,
-		'data_length' : SZUSHORT,
-		'reserved' : SZUSHORT,
-		SYMBOLIC_LINK : {
-			'substitute_name_offset' : SZUSHORT,
-			'substitute_name_length' : SZUSHORT,
-			'print_name_offset' : SZUSHORT,
-			'print_name_length' : SZUSHORT,
-			'flags' : SZULONG,
-			'buffer' : u'',
-			'pkeys' : [
-				'substitute_name_offset',
+	data = {'tag': struct.unpack('<I', buf[:4])[0],
+			'data_length': struct.unpack('<H', buf[4:6])[0],
+			'reserved': struct.unpack('<H', buf[6:8])[0]}
+	buf = buf[8:]
+
+	if data['tag'] in (IO_REPARSE_TAG_MOUNT_POINT, IO_REPARSE_TAG_SYMLINK):
+		keys = ['substitute_name_offset',
 				'substitute_name_length',
 				'print_name_offset',
-				'print_name_length',
-				'flags',
-			]
-		},
-		MOUNTPOINT : {
-			'substitute_name_offset' : SZUSHORT,
-			'substitute_name_length' : SZUSHORT,
-			'print_name_offset' : SZUSHORT,
-			'print_name_length' : SZUSHORT,
-			'buffer' : u'',
-			'pkeys' : [
-				'substitute_name_offset',
-				'substitute_name_length',
-				'print_name_offset',
-				'print_name_length',
-			]
-		},
-		GENERIC : {
-			'pkeys' : [],
-			'buffer': ''
-		}
-	}
+				'print_name_length']
+		if data['tag'] == IO_REPARSE_TAG_SYMLINK:
+			keys.append('flags')
 
-	# Header stuff
-	buffer['tag'] = original[:SZULONG]
-	buffer['data_length'] = original[SZULONG:SZUSHORT]
-	buffer['reserved'] = original[SZULONG+SZUSHORT:SZUSHORT]
-	original = original[8:]
+		# Parsing
+		for k in keys:
+			if k == 'flags':
+				fmt, sz = '<I', 4
+			else:
+				fmt, sz = '<H', 2
+			data[k] = struct.unpack(fmt, buf[:sz])[0]
+			buf = buf[sz:]
 
-	# Parsing
-	k = reparse_type
-	for c in buffer[k]['pkeys']:
-		if type(buffer[k][c]) == int:
-			sz = buffer[k][c]
-			bytes = original[:sz]
-			buffer[k][c] = 0
-			for b in bytes:
-				n = ord(b)
-				if n:
-					buffer[k][c] += n
-			original = original[sz:]
+	# Using the offset and lengths grabbed, we'll set the buffer.
+	data['buffer'] = buf
 
-	# Using the offset and length's grabbed, we'll set the buffer.
-	buffer[k]['buffer'] = original
-	return buffer
+	return data
 
 
 def readlink(path):
@@ -603,24 +652,31 @@ def readlink(path):
 		createfilefn = CreateFileW
 	else:
 		createfilefn = CreateFile
+	# FILE_FLAG_OPEN_REPARSE_POINT alone is not enough if 'path'
+	# is a symbolic link to a directory or a NTFS junction.
+	# We need to set FILE_FLAG_BACKUP_SEMANTICS as well.
+	# See https://docs.microsoft.com/en-us/windows/desktop/api/fileapi/nf-fileapi-createfilea
 	handle = createfilefn(path, GENERIC_READ, 0, None, OPEN_EXISTING,
-						  FILE_FLAG_OPEN_REPARSE_POINT, 0)
+						  FILE_FLAG_BACKUP_SEMANTICS | FILE_FLAG_OPEN_REPARSE_POINT, 0)
 
 	# MAXIMUM_REPARSE_DATA_BUFFER_SIZE = 16384 = (16 * 1024)
-	buffer = DeviceIoControl(handle, FSCTL_GET_REPARSE_POINT, None, 16 * 1024)
+	buf = DeviceIoControl(handle, FSCTL_GET_REPARSE_POINT, None, 16 * 1024)
 	# Above will return an ugly string (byte array), so we'll need to parse it.
 
 	# But first, we'll close the handle to our file so we're not locking it anymore.
 	CloseHandle(handle)
 
 	# Minimum possible length (assuming that the length is bigger than 0)
-	if len(buffer) < 9:
+	if len(buf) < 9:
 		return type(path)()
 	# Parse and return our result.
-	result = parse_reparse_buffer(buffer)
-	offset = result[SYMBOLIC_LINK]['substitute_name_offset']
-	ending = offset + result[SYMBOLIC_LINK]['substitute_name_length']
-	rpath = result[SYMBOLIC_LINK]['buffer'][offset:ending].decode('UTF-16-LE')
+	result = parse_reparse_buffer(buf)
+	if result['tag'] in (IO_REPARSE_TAG_MOUNT_POINT, IO_REPARSE_TAG_SYMLINK):
+		offset = result['substitute_name_offset']
+		ending = offset + result['substitute_name_length']
+		rpath = result['buffer'][offset:ending].decode('UTF-16-LE')
+	else:
+		rpath = result['buffer']
 	if len(rpath) > 4 and rpath[0:4] == '\\??\\':
 		rpath = rpath[4:]
 	return rpath
@@ -638,6 +694,141 @@ def relpath(path, start):
 		return os.path.sep.join([".."] * (len(start) - len(path)))
 
 
+def safe_glob(pathname):
+	"""
+	Return a list of paths matching a pathname pattern.
+
+	The pattern may contain simple shell-style wildcards a la
+	fnmatch. However, unlike fnmatch, filenames starting with a
+	dot are special cases that are not matched by '*' and '?'
+	patterns.
+
+	Like fnmatch.glob, but suppresses re.compile errors by escaping
+	uncompilable path components.
+
+	See https://bugs.python.org/issue738361
+
+	"""
+	return list(safe_iglob(pathname))
+
+
+def safe_iglob(pathname):
+	"""
+	Return an iterator which yields the paths matching a pathname pattern.
+
+	The pattern may contain simple shell-style wildcards a la
+	fnmatch. However, unlike fnmatch, filenames starting with a
+	dot are special cases that are not matched by '*' and '?'
+	patterns.
+
+	Like fnmatch.iglob, but suppresses re.compile errors by escaping
+	uncompilable path components.
+	
+	See https://bugs.python.org/issue738361
+
+	"""
+	dirname, basename = os.path.split(pathname)
+	if not glob.has_magic(pathname):
+		if basename:
+			if os.path.lexists(pathname):
+				yield pathname
+		else:
+			# Patterns ending with a slash should match only directories
+			if os.path.isdir(dirname):
+				yield pathname
+		return
+	if not dirname:
+		for name in safe_glob1(os.curdir, basename):
+			yield name
+		return
+	# `os.path.split()` returns the argument itself as a dirname if it is a
+	# drive or UNC path.  Prevent an infinite recursion if a drive or UNC path
+	# contains magic characters (i.e. r'\\?\C:').
+	if dirname != pathname and glob.has_magic(dirname):
+		dirs = safe_iglob(dirname)
+	else:
+		dirs = [dirname]
+	if glob.has_magic(basename):
+		glob_in_dir = safe_glob1
+	else:
+		glob_in_dir = glob.glob0
+	for dirname in dirs:
+		for name in glob_in_dir(dirname, basename):
+			yield os.path.join(dirname, name)
+
+
+def safe_glob1(dirname, pattern):
+	if not dirname:
+		dirname = os.curdir
+	if isinstance(pattern, unicode) and not isinstance(dirname, unicode):
+		dirname = unicode(dirname, sys.getfilesystemencoding() or
+								   sys.getdefaultencoding())
+	try:
+		names = os.listdir(dirname)
+	except os.error:
+		return []
+	if pattern[0] != '.':
+		names = filter(lambda x: x[0] != '.', names)
+	return safe_shell_filter(names, pattern)
+
+
+def safe_shell_filter(names, pat):
+	"""
+	Return the subset of the list NAMES that match PAT
+
+	Like fnmatch.filter, but suppresses re.compile errors by escaping
+	uncompilable path components.
+	
+	See https://bugs.python.org/issue738361
+	
+	"""
+	import posixpath
+	result = []
+	pat = os.path.normcase(pat)
+	try:
+		re_pat = _cache[pat]
+	except KeyError:
+		res = safe_translate(pat)
+		if len(_cache) >= _MAXCACHE:
+			_cache.clear()
+		_cache[pat] = re_pat = re.compile(res)
+	match = re_pat.match
+	if os.path is posixpath:
+		# normcase on posix is NOP. Optimize it away from the loop.
+		for name in names:
+			if match(name):
+				result.append(name)
+	else:
+		for name in names:
+			if match(os.path.normcase(name)):
+				result.append(name)
+	return result
+
+
+def safe_translate(pat):
+	"""
+	Translate a shell PATTERN to a regular expression.
+
+	Like fnmatch.translate, but suppresses re.compile errors by escaping
+	uncompilable path components.
+	
+	See https://bugs.python.org/issue738361
+	
+	"""
+	if isinstance(getattr(os.path, "altsep", None), basestring):
+		# Normalize path separators
+		pat = pat.replace(os.path.altsep, os.path.sep)
+	components = pat.split(os.path.sep)
+	for i, component in enumerate(components):
+		translated = fnmatch.translate(component)
+		try:
+			re.compile(translated)
+		except re.error:
+			translated = re.escape(component)
+		components[i] = translated
+	return re.escape(os.path.sep).join(components)
+
+
 def waccess(path, mode):
 	""" Test access to path """
 	if mode & os.R_OK:
@@ -723,6 +914,75 @@ def whereis(names, bin=True, bin_paths=None, man=True, man_paths=None, src=True,
 	return result
 
 
+class FileLock(object):
+
+	if sys.platform == "win32":
+		_exception_cls = pywintypes.error
+	else:
+		_exception_cls = IOError
+
+	def __init__(self, file_, exclusive=False, blocking=False):
+		self._file = file_
+		self.exclusive = exclusive
+		self.blocking = blocking
+		self.lock()
+
+	def __enter__(self):
+		return self
+
+	def __exit__(self, etype, value, traceback):
+		self.unlock()
+
+	def lock(self):
+		if sys.platform == "win32":
+			mode = 0
+			if self.exclusive:
+				mode |= win32con.LOCKFILE_EXCLUSIVE_LOCK
+			if not self.blocking:
+				mode |= win32con.LOCKFILE_FAIL_IMMEDIATELY
+			self._handle = win32file._get_osfhandle(self._file.fileno())
+			self._overlapped = pywintypes.OVERLAPPED()
+			fn = win32file.LockFileEx
+			args = (self._handle, mode, 0, -0x10000, self._overlapped)
+		else:
+			if self.exclusive:
+				op = fcntl.LOCK_EX
+			else:
+				op = fcntl.LOCK_SH
+			if not self.blocking:
+				op |= fcntl.LOCK_NB
+			fn = fcntl.flock
+			args = (self._file.fileno(), op)
+		self._call(fn, args, FileLock.LockingError)
+
+	def unlock(self):
+		if self._file.closed:
+			return
+		if sys.platform == "win32":
+			fn = win32file.UnlockFileEx
+			args = (self._handle, 0, -0x10000, self._overlapped)
+		else:
+			fn = fcntl.flock
+			args = (self._file.fileno(), fcntl.LOCK_UN)
+		self._call(fn, args, FileLock.UnlockingError)
+
+	@staticmethod
+	def _call(fn, args, exception_cls):
+		try:
+			fn(*args)
+		except FileLock._exception_cls, exception:
+			raise exception_cls(*exception.args)
+
+	class Error(Exception):
+		pass
+
+	class LockingError(Error):
+		pass
+
+	class UnlockingError(Error):
+		pass
+
+
 if sys.platform == "win32" and sys.getwindowsversion() >= (6, ):
 	class win64_disable_file_system_redirection:
 
