@@ -7,7 +7,11 @@ import threading
 import types
 
 from colormath import specialpow
-from wxfixes import wx
+from wxfixes import wx, GenButton, PlateButton, get_dialogs
+
+import wx.grid
+from lib.agw.gradientbutton import GradientButton
+import floatspin
 
 
 def AdjustMinMax(self, minvalue=0.0, maxvalue=1.0):
@@ -54,11 +58,13 @@ def IsBW(self):
 	non-equal RGB triplet).
 	
 	"""
+	triplet = set()
 	for i, byte in enumerate(self.GetDataBuffer()):
-			if i % 3 == 2:
-					if byte != prev_byte:
-							return False
-			prev_byte = byte
+		triplet.add(byte)
+		if i % 3 == 2:
+			if len(triplet) != 1:
+				return False
+			triplet = set()
 	return True
 
 wx.Image.IsBW = IsBW
@@ -146,21 +152,43 @@ def SetSaneGeometry(self, x=None, y=None, w=None, h=None):
 	# is completely outside the client area of all displays
 	display_client_rect = self.GetDisplay().ClientArea 
 	if sys.platform not in ("darwin", "win32"): # Linux
-		safety_margin = 40
+		if os.getenv("XDG_SESSION_TYPE") == "wayland":
+			# Client-side decorations
+			safety_margin = 0
+		else:
+			# Assume server-side decorations
+			safety_margin = 40
 	else:
 		safety_margin = 20
 	if not None in (w, h):
 		# Set given size, but resize if needed to fit inside client area
-		min_w, min_h = self.MinSize
-		self.SetSize((max(min(display_client_rect[2], w), min_w), 
-					  max(min(display_client_rect[3] - safety_margin, h),
-						  min_h)))
+		if hasattr(self, "MinClientSize"):
+			min_w, min_h = self.MinClientSize
+		else:
+			min_w, min_h = self.WindowToClientSize(self.MinSize)
+		border_lr = self.Size[0] - self.ClientSize[0]
+		border_tb = self.Size[1] - self.ClientSize[1]
+		self.ClientSize = (min(display_client_rect[2] - border_lr, max(w, min_w)), 
+						   min(display_client_rect[3] - border_tb -
+							   safety_margin, max(h, min_h)))
 	if not None in (x, y):
-		if not display_client_rect.ContainsXY(x, y) or \
-		   not display_client_rect.ContainsRect((x, y, 100, 100)):
-			# If outside client area or near the borders,
-			# move to leftmost / topmost coordinates of client area
-			self.SetPosition(display_client_rect[0:2])
+		if (not display_client_rect.ContainsXY(x, y) or
+			not display_client_rect.ContainsRect((x, y,
+												  self.Size[0], self.Size[1]))):
+			# If outside client area, move into client area
+			xy = [x, y]
+			for i, pos in enumerate([xy,
+									 (x + self.Size[0], y + self.Size[1])]):
+				for j in xrange(2):
+					if (pos[j] > display_client_rect[j] +
+								 display_client_rect[2 + j] or
+						pos[j] < display_client_rect[j]):
+						if i:
+							xy[j] = (display_client_rect[j] +
+									 display_client_rect[2 + j] - self.Size[j])
+						else:
+							xy[j] = display_client_rect[j]
+			self.SetPosition(tuple(xy))
 
 wx.Window.SetSaneGeometry = SetSaneGeometry
 
@@ -519,15 +547,22 @@ class ThreadedCallLater(ThreadedTimer):
 
 class BetterWindowDisabler(object):
 	
-	""" Also disables child windows instead of only top level windows. This is
+	"""
+	Also disables child windows instead of only top level windows. This is
 	actually needed under Mac OS X where disabling a top level window will
-	not prevent interaction with its children. """
+	not prevent interaction with its children.
+	
+	If toplevelparent is given, disable only this window and its child windows.
+	
+	"""
 	
 	windows = set()
 
-	def __init__(self, skip=None):
+	def __init__(self, skip=None, toplevelparent=None, include_menus=False):
 		self._windows = []
 		self.skip = skip
+		self.toplevelparent = toplevelparent
+		self.include_menus = include_menus
 		self.disable()
 
 	def __del__(self):
@@ -542,18 +577,46 @@ class BetterWindowDisabler(object):
 			if skip:
 				if not isinstance(skip, (list, tuple)):
 					skip = [skip]
-			toplevel = list(wx.GetTopLevelWindows())
+			if self.toplevelparent:
+				toplevel = [self.toplevelparent]
+			else:
+				toplevel = list(wx.GetTopLevelWindows())
 			for w in toplevel:
 				if (w and w not in skip and "Inspection" not in "%s" % w and
 					w not in BetterWindowDisabler.windows):
 					self._windows.append(w)
+					# Selectively add children to our list of handled
+					# windows. This prevents a segfault with wxPython 4
+					# under macOS where GetAllChildren includes sub-controls
+					# of controls, like scrollbars etc.
 					for child in w.GetAllChildren(skip + toplevel):
-						if (child and not isinstance(child, (wx.Panel,
-															 wx.PyPanel)) and
+						if (child and isinstance(child, (wx.BitmapButton,
+														 wx.Button,
+														 wx.CheckBox,
+														 wx.Choice,
+														 wx.ComboBox,
+														 wx.ListBox,
+														 wx.ListCtrl,
+														 wx.RadioButton,
+														 wx.SpinCtrl,
+														 wx.Slider,
+														 wx.StaticText,
+														 wx.TextCtrl,
+														 wx.grid.Grid,
+														 floatspin.FloatSpin,
+														 GenButton,
+														 GradientButton,
+														 PlateButton)) and
 							child not in BetterWindowDisabler.windows):
 							# Don't disable panels, this can have weird side
 							# effects for contained controls
 							self._windows.append(child)
+					if self.include_menus:
+						menubar = w.GetMenuBar()
+						if menubar:
+							for menu, label in menubar.GetMenus():
+								for item in menu.GetMenuItems():
+									self._windows.append(item)
 			def Enable(w, enable=True):
 				w._BetterWindowDisabler_enabled = enable
 			def Disable(w):
@@ -561,9 +624,10 @@ class BetterWindowDisabler(object):
 			for w in reversed(self._windows):
 				BetterWindowDisabler.windows.add(w)
 				enabled = w.IsEnabled()
-				w.Disable()
-				w._BetterWindowDisabler_Disable = w.Disable
-				w.Disable = types.MethodType(Disable, w, type(w))
+				w.Enable(False)
+				if hasattr(w, "Disable"):
+					w._BetterWindowDisabler_Disable = w.Disable
+					w.Disable = types.MethodType(Disable, w, type(w))
 				w._BetterWindowDisabler_Enable = w.Enable
 				w.Enable = types.MethodType(Enable, w, type(w))
 				w.Enable(enabled)
@@ -592,6 +656,113 @@ class CustomGridCellEvent(CustomEvent):
 		return self.Col
 
 
+class PopupMenu(object):
+
+	"""
+	A collection of menus that has a wx.MenuBar-like interface
+	
+	"""
+
+	def __init__(self, parent):
+		self.Parent = parent
+		self.TopLevelParent = parent.TopLevelParent
+		self._menus = []
+		self._enabledtop = {}
+
+	def Append(self, menu, title):
+		self._menus.append((menu, title))
+
+	def EnableTop(self, pos, enable=True):
+		self._enabledtop[pos] = enable
+
+	def FindItemById(self, id):
+		for menu, label in self._menus:
+			item = menu.FindItemById(id)
+			if item:
+				return item
+
+	def FindMenu(self, title):
+		for i, (menu, label) in enumerate(self._menus):
+			if title == label:
+				return i
+		return wx.NOT_FOUND
+
+	def GetMenu(self, index):
+		return self._menus[index][0]
+
+	def GetMenuCount(self):
+		return len(self._menus)
+
+	def GetMenus(self):
+		return list(self._menus)
+
+	def IsEnabledTop(self, pos):
+		return self._enabledtop.get(pos, True)
+
+	def SetMenuLabel(self, pos, label):
+		self._menus[pos] = (self._menus[pos][0], label)
+
+	def SetMenus(self, menus):
+		self._menus = []
+		for menu, label in menus:
+			self.Append((menu, label))
+
+	def bind_keys(self):
+		if sys.platform == "darwin":
+			accels = self.get_accelerator_entries()
+			self.TopLevelParent.SetAcceleratorTable(wx.AcceleratorTable(accels()))
+		else:
+			self.TopLevelParent.Bind(wx.EVT_CHAR_HOOK, self.key_handler)
+
+	def get_accelerator_entries(self):
+		accels = []
+		for menu, label in self._menus:
+			for item in menu.MenuItems:
+				accel = item.Accel
+				if accel:
+					accel = wx.AcceleratorEntry(accel.Flags, accel.KeyCode, accel.Command, item)
+					accels.append(accel)
+		return accels
+
+	def key_handler(self, event):
+		""" Handle accelerator keys """
+		keycode = event.KeyCode
+		flags = wx.ACCEL_NORMAL
+		for key in ("ALT", "CMD", "CTRL", "SHIFT"):
+			if wx.GetKeyState(getattr(wx, "WXK_" + key.replace("CTRL", "CONTROL"), -1)):
+				flags |= getattr(wx, "ACCEL_" + key.upper())
+		for menu, label in self._menus:
+			for item in menu.MenuItems:
+				accel = item.Accel
+				if accel and accel.KeyCode == keycode and accel.Flags == flags:
+					event = wx.CommandEvent(wx.wxEVT_COMMAND_MENU_SELECTED)
+					event.Id = item.Id
+					if item.Kind == wx.ITEM_RADIO:
+						event.SetInt(1)
+					elif item.Kind == wx.ITEM_CHECK:
+						event.SetInt(int(not item.Checked))
+					self.TopLevelParent.ProcessEvent(event)
+
+	def popup(self):
+		""" Popup the list of menus (with actual menus as submenus) """
+
+		top_menu = wx.Menu()
+
+		for menu, label in self._menus:
+			top_menu.AppendSubMenu(menu, label)
+
+		self.Parent.PopupMenu(top_menu)
+
+		# Delete menuitems (not submenus)
+		for item in top_menu.MenuItems:
+			top_menu.Delete(item)
+
+		# Now we can safely destroy the menu without affecting submenus
+		top_menu.Destroy()
+
+	Menus = property(GetMenus, SetMenus)
+
+
 class FileDrop(wx.FileDropTarget):
 
 	def __init__(self, drophandlers=None):
@@ -602,6 +773,13 @@ class FileDrop(wx.FileDropTarget):
 		self.unsupported_handler = None
 
 	def OnDropFiles(self, x, y, filenames):
+		dialogs = get_dialogs()
+		interactable = (not hasattr(self, "parent") or
+						(self.parent.Enabled and
+						 (not dialogs or self.parent in dialogs)))
+		if not interactable:
+			wx.Bell()
+			return False
 		self._files = []
 		self._filenames = filenames
 
