@@ -8,6 +8,11 @@ import tempfile
 
 from meta import name as appname, wx_minversion
 
+if os.getenv("GTK_CSD", "0") != "0":
+	# Work-around double window decorations, see
+	# https://hub.displaycal.net/issue/17331/
+	os.environ["GTK_CSD"] = "0"
+
 # wxversion will be removed in Phoenix
 try:
 	import wxversion
@@ -15,7 +20,10 @@ except ImportError:
 	pass
 else:
 	if not getattr(sys, "frozen", False) and not "wx" in sys.modules:
-		wxversion.ensureMinimal("%i.%i" % wx_minversion[:2])
+		try:
+			wxversion.select(["4.0", "3.0", "%i.%i.%i" % wx_minversion[:3]])
+		except wxversion.VersionError:
+			pass
 import wx
 if wx.VERSION < wx_minversion:
 	app = wx.GetApp() or wx.App(0)
@@ -516,6 +524,57 @@ if not hasattr(wx.Sizer, "GetItemIndex"):
 	wx.Sizer.GetItemIndex = GetItemIndex
 
 
+def set_maxsize(window, maxsize):
+	window.MaxSize = maxsize
+
+
+if os.getenv("XDG_SESSION_TYPE") == "wayland":
+	# Fix erroneous extra spacing around window contents under Wayland
+
+	is_first_toplevel_window = True
+
+	def fix_wayland_window_size(window):
+		global is_first_toplevel_window
+		if is_first_toplevel_window:
+			# Not needed for first toplevel window
+			is_first_toplevel_window = False
+		else:
+			window.MaxSize = window.Size
+			wx.CallAfter(set_maxsize, window, (-1, -1))
+
+	def TopLevelWindow_Show(self, show=True):
+		if show and not self.IsShown():
+			fix_wayland_window_size(self)
+		return _TopLevelWindow_Show(self, show)
+
+	_TopLevelWindow_Show = wx.TopLevelWindow.Show
+	wx.TopLevelWindow.Show = TopLevelWindow_Show
+
+	# ~ def Dialog_ShowModal(self):
+		# ~ fix_wayland_window_size(self)
+		# ~ return _Dialog_ShowModal(self)
+
+	# ~ _Dialog_ShowModal = wx.Dialog.ShowModal
+	# ~ wx.Dialog.ShowModal = Dialog_ShowModal
+
+	def Sizer_SetSizeHints(self, window):
+		if isinstance(window, wx.Dialog):
+			window.MaxSize = (-1, -1)
+		_Sizer_SetSizeHints(self, window)
+
+	_Sizer_SetSizeHints = wx.Sizer.SetSizeHints
+	wx.Sizer.SetSizeHints = Sizer_SetSizeHints
+
+	def Sizer_Layout(self):
+		_Sizer_Layout(self)
+		window = self.GetContainingWindow()
+		if isinstance(window, wx.Dialog):
+			window.MaxSize = window.Size
+
+	_Sizer_Layout = wx.Sizer.Layout
+	wx.Sizer.Layout = Sizer_Layout
+
+
 if sys.platform == "darwin":
 	# wxMac seems to loose foreground color of StaticText
 	# when enabled again
@@ -578,36 +637,43 @@ def SetToolTipString(self, string):
 wx.Window.SetToolTipString = SetToolTipString
 
 
+def _adjust_sizer_args_scaling_for_appdpi(*args, **kwargs):
+	from config import get_default_dpi, getcfg
+	scale = getcfg("app.dpi") / get_default_dpi()
+	if scale > 1:
+		args = list(args)
+		if kwargs.get("border"):
+			kwargs["border"] = int(round(kwargs["border"] * scale))
+		elif not "border" in kwargs:
+			if (isinstance(args[0],
+						   (tuple, wx.Size, wx.Sizer, wx.Window)) and
+				len(args) > 3 and args[3]):
+				# item, proportion, flag, border, ...
+				args[3] = int(round(args[3] * scale))
+			elif isinstance(args[0], int) and len(args) > 4 and args[4]:
+				# width, height, proportion, flag, border, ...
+				args[4] = int(round(args[4] * scale))
+		if args and isinstance(args[0], (tuple, wx.Size)):
+			spacer = list(args[0])
+			##print spacer, '->',
+			for i, dimension in enumerate(spacer):
+				if dimension > 0:
+					spacer[i] = int(round(dimension * scale))
+			##print spacer
+			args = [tuple(spacer)] + args[1:]
+	return args, kwargs
+
 wx._Sizer = wx.Sizer
 
 class Sizer(wx._Sizer):
 
 	def Add(self, *args, **kwargs):
-		from config import get_default_dpi, getcfg
-		scale = getcfg("app.dpi") / get_default_dpi()
-		if isinstance(args[0], int):
-			args = list(args)
-			index = args.pop(0)
-			args = tuple(args)
-		else:
-			index = None
-		if scale > 1:
-			if kwargs.get("border"):
-				kwargs["border"] = int(round(kwargs["border"] * scale))
-			if args and isinstance(args[0], tuple):
-				spacer = list(args[0])
-				##print spacer, '->',
-				for i, dimension in enumerate(spacer):
-					if dimension > 0:
-						spacer[i] = int(round(dimension * scale))
-				##print spacer
-				args = (tuple(spacer), ) + args[1:]
-		if index is None:
-			return wx._Sizer.Add(self, *args, **kwargs)
-		else:
-			return wx._Sizer.Insert(self, index, *args, **kwargs)
+		args, kwargs = _adjust_sizer_args_scaling_for_appdpi(*args, **kwargs)
+		return wx._Sizer.Add(self, *args, **kwargs)
 
-	Insert = Add
+	def Insert(self, index, *args, **kwargs):
+		args, kwargs = _adjust_sizer_args_scaling_for_appdpi(*args, **kwargs)
+		return wx._Sizer.Insert(self, index, *args, **kwargs)
 
 wx.Sizer = Sizer
 
@@ -617,7 +683,7 @@ wx._BoxSizer = wx.BoxSizer
 class BoxSizer(wx._BoxSizer):
 
 	Add = Sizer.__dict__["Add"]
-	Insert = Add
+	Insert = Sizer.__dict__["Insert"]
 
 wx.BoxSizer = BoxSizer
 
@@ -637,7 +703,7 @@ class GridSizer(wx._GridSizer):
 		super(GridSizer, self).__init__(rows, cols, vgap, hgap)
 
 	Add = Sizer.__dict__["Add"]
-	Insert = Add
+	Insert = Sizer.__dict__["Insert"]
 
 wx.GridSizer = GridSizer
 
@@ -657,7 +723,7 @@ class FlexGridSizer(wx._FlexGridSizer):
 		super(FlexGridSizer, self).__init__(rows, cols, vgap, hgap)
 
 	Add = Sizer.__dict__["Add"]
-	Insert = Add
+	Insert = Sizer.__dict__["Insert"]
 
 wx.FlexGridSizer = FlexGridSizer
 
@@ -904,6 +970,15 @@ def set_bitmap_labels(btn, disabled=True, focus=None, pressed=True):
 			btn.SetBitmapSelected(bmp)
 
 
+def get_dialogs(modal=False):
+	""" If there are any dialogs open, return them """
+	return filter(lambda window: window and
+								 isinstance(window, wx.Dialog) and
+								 window.IsShown() and
+								 (not modal or window.IsModal()),
+				  wx.GetTopLevelWindows())
+
+
 # wx.DirDialog and wx.FileDialog are normally not returned by
 # wx.GetTopLevelWindows, do some trickery to work around
 
@@ -912,8 +987,8 @@ _FileDialog = wx.FileDialog
 
 class PathDialogBase(wx.Dialog):
 
-	def __init__(self, name):
-		wx.Dialog.__init__(self, None, -1, name=name)
+	def __init__(self, parent, name):
+		wx.Dialog.__init__(self, parent, -1, name=name)
 		self._ismodal = False
 		self._isshown = False
 		self.Bind(wx.EVT_WINDOW_DESTROY, self.OnDestroy)
@@ -946,13 +1021,13 @@ class PathDialogBase(wx.Dialog):
 class DirDialog(PathDialogBase):
 
 	def __init__(self, *args, **kwargs):
-		PathDialogBase.__init__(self, name="dirdialog")
+		PathDialogBase.__init__(self, args[0], name="dirdialog")
 		self.filedialog = _DirDialog(*args, **kwargs)
 
 class FileDialog(PathDialogBase):
 
 	def __init__(self, *args, **kwargs):
-		PathDialogBase.__init__(self, name="filedialog")
+		PathDialogBase.__init__(self, args[0], name="filedialog")
 		self.filedialog = _FileDialog(*args, **kwargs)
 
 wx.DirDialog = DirDialog
@@ -1288,10 +1363,23 @@ class PlateButton(platebtn.PlateButton):
 	_reallyenabled = True
 
 	def __init__(self, *args, **kwargs):
+		from config import get_default_dpi, getcfg
+		self.dpiscale = getcfg("app.dpi") / get_default_dpi()
 		platebtn.PlateButton.__init__(self, *args, **kwargs)
+		self._bmp["hilite"] = None
+		if sys.platform == "darwin":
+			# Use Sierra-like color scheme
+			from wxaddons import gamma_encode
+			color = wx.Colour(*gamma_encode(0, 105, 217))
+		else:
+			color = wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHT)
+		self._color['hlight'] = color
 		self.Unbind(wx.EVT_LEAVE_WINDOW)
 		self.Bind(wx.EVT_LEAVE_WINDOW,
 				  lambda evt: wx.CallLater(80, self.__LeaveWindow))
+		if ("gtk2" in wx.PlatformInfo and
+			not isinstance(self.TopLevelParent, wx.Dialog)):
+			self.BackgroundColour = self.Parent.BackgroundColour
 
 	def DoGetBestSize(self):
 		"""Calculate the best size of the button
@@ -1299,9 +1387,9 @@ class PlateButton(platebtn.PlateButton):
 		:return: :class:`Size`
 
 		"""
-		# A liitle more padding left + right
-		width = 24
-		height = 6
+		# A little more padding left + right
+		width = 6 * self.dpiscale
+		height = 6 * self.dpiscale
 		if self.Label:
 			# NOTE: Should measure with a GraphicsContext to get right
 			#       size, but due to random segfaults on linux special
@@ -1312,16 +1400,17 @@ class PlateButton(platebtn.PlateButton):
 			
 		if self._bmp['enable'] is not None:
 			bsize = self._bmp['enable'].Size
-			width += (bsize[0] + 10)
+			width += (bsize[0] + 10 * self.dpiscale)
 			if height <= bsize[1]:
-				height = bsize[1] + 6
+				height = bsize[1] + 6 * self.dpiscale
 			else:
+				# No DPI scaling needed here
 				height += 3
 		else:
-			width += 10
+			width += 10 * self.dpiscale
 
 		if self._menu is not None or self._style & platebtn.PB_STYLE_DROPARROW:
-			width += 12
+			width += 12 * self.dpiscale
 
 		best = wx.Size(width, height)
 		self.CacheBestSize(best)
@@ -1335,7 +1424,16 @@ class PlateButton(platebtn.PlateButton):
 		"""
 		return self._bmp["enable"]
 
-	def __DrawBitmap(self, gc):
+
+	def SetBitmapHover(self, bmp):
+		"""Set the bitmap displayed in the button
+
+		:param `bmp`: :class:`wx.Bitmap`
+
+		"""
+		self._bmp['hilite'] = bmp
+
+	def __DrawBitmap(self, gc, interactable):
 		"""Draw the bitmap if one has been set
 
 		:param GCDC `gc`: :class:`GCDC` to draw with
@@ -1343,11 +1441,15 @@ class PlateButton(platebtn.PlateButton):
 
 		"""
 		if self.IsEnabled():
-			bmp = self._bmp['enable']
+			if self._state['cur'] in (platebtn.PLATE_HIGHLIGHT,
+									  platebtn.PLATE_PRESSED) and interactable:
+				bmp = self._bmp['hilite']
+			else:
+				bmp = self._bmp['enable']
 		else:
 			bmp = self._bmp['disable']
 
-		xpos = 12
+		xpos = 4 * self.dpiscale
 		if bmp is not None and bmp.IsOk():
 			bw, bh = bmp.GetSize()
 			ypos = (self.GetSize()[1] - bh) // 2
@@ -1365,7 +1467,7 @@ class PlateButton(platebtn.PlateButton):
 			dc = wx.PaintDC(self)
 		else:
 			dc = wx.AutoBufferedPaintDCFactory(self)
-			dc.SetBackground(wx.Brush(self.Parent.BackgroundColour))
+			dc.SetBackground(wx.TRANSPARENT_BRUSH)
 			dc.Clear()
 
 		gc = wx.GCDC(dc)
@@ -1376,11 +1478,12 @@ class PlateButton(platebtn.PlateButton):
 
 		# Calc Object Positions
 		width, height = self.GetSize()
-		# XXX: Using self.GetTextextent instead of gc.GetTextExtent
-		# seems to fix sporadic segfaults with wxPython Phoenix under Windows.
-		# TODO: Figure out why this is the case.
-		tw, th = self.GetTextExtent(self.Label)
-		txt_y = max((height - th) // 2, 1)
+		# Note: Using self.GetTextExtent instead of gc.GetTextExtent seems
+		# to fix sporadic segfaults with wxPython Phoenix up to 4.0.0a2
+		# under Windows (fixed in 4.0.0a3), but self.GetTextExtent is NOT
+		# an equivalent replacement for gc.GetTextExtent.
+		tw, th = gc.GetTextExtent(self.Label)
+		txt_y = max((height - th) // 2 - 1, 1)
 
 		# The background needs some help to look transparent on
 		# on Gtk and Windows
@@ -1391,7 +1494,16 @@ class PlateButton(platebtn.PlateButton):
 
 		gc.SetBrush(wx.TRANSPARENT_BRUSH)
 
-		if self._state['cur'] == platebtn.PLATE_HIGHLIGHT and self.IsEnabled():
+		space = 4 * self.dpiscale
+
+		# Our IsEnabled() override only deals with explit enabled/disabled
+		# state, while we also need to deal with interaction state
+		# (button enabled and interactable, eg. no modal dialog shown that
+		# would prevent interaction)
+		dialogs = get_dialogs()
+		interactable = not dialogs or self.TopLevelParent in dialogs
+		if (self._state['cur'] == platebtn.PLATE_HIGHLIGHT and
+			self.IsEnabled() and interactable):
 			gc.SetTextForeground(self._color['htxt'])
 			gc.SetPen(wx.TRANSPARENT_PEN)
 			self.__DrawHighlight(gc, width, height)
@@ -1402,8 +1514,8 @@ class PlateButton(platebtn.PlateButton):
 			gc.SetPen(pen)
 
 			self.__DrawHighlight(gc, width, height)
-			txt_x = self.__DrawBitmap(gc)
-			t_x = max((width - tw - (txt_x + 8)) // 2, txt_x + 8)
+			txt_x = self.__DrawBitmap(gc, interactable)
+			t_x = max((width - tw - (txt_x + space)) // 2, txt_x + space)
 			gc.DrawText(self.Label, t_x, txt_y)
 			self.__DrawDropArrow(gc, width - 10, (height // 2) - 2)
 
@@ -1416,8 +1528,8 @@ class PlateButton(platebtn.PlateButton):
 
 		# Draw bitmap and text
 		if self._state['cur'] != platebtn.PLATE_PRESSED or not self.IsEnabled():
-			txt_x = self.__DrawBitmap(gc)
-			t_x = max((width - tw - (txt_x + 8)) // 2, txt_x + 8)
+			txt_x = self.__DrawBitmap(gc, interactable)
+			t_x = max((width - tw - (txt_x + space)) // 2, txt_x + space)
 			gc.DrawText(self.Label, t_x, txt_y)
 			self.__DrawDropArrow(gc, width - 10, (height // 2) - 2)
 
@@ -1463,7 +1575,7 @@ class TempXmlResource(object):
 
 	def __init__(self, xmlpath):
 		from config import get_default_dpi, getcfg
-		scale = getcfg("app.dpi") / get_default_dpi()
+		scale = max(getcfg("app.dpi") / get_default_dpi(), 1)
 		if scale > 1 or "gtk3" in wx.PlatformInfo:
 			if not TempXmlResource._temp:
 				try:
@@ -1519,8 +1631,9 @@ class TempXmlResource(object):
 		return getattr(self.res, name)
 
 
-class ThemedGenBitmapTextButton(ThemedGenButton, _GenBitmapTextButton):
-	"""A themed generic bitmapped button with text label"""
+class GenBitmapTextButton(GenButton, _GenBitmapTextButton):
+	"""A generic bitmapped button with text label"""
+
 	def __init__(self, parent, id=-1, bitmap=wx.NullBitmap, label='',
 				 pos=wx.DefaultPosition, size=wx.DefaultSize, style=0,
 				 validator=wx.DefaultValidator, name="genbutton"):
@@ -1564,6 +1677,10 @@ class ThemedGenBitmapTextButton(ThemedGenButton, _GenBitmapTextButton):
 		dc.DrawText(label, pos_x + dx+bw, (height-th)/2+dy)      # draw the text
 
 
+class ThemedGenBitmapTextButton(ThemedGenButton, GenBitmapTextButton):
+	"""A themed generic bitmapped button with text label"""
+	pass
+
 class BitmapWithThemedButton(wx.BoxSizer):
 
 	def __init__(self, parent, id=-1, bitmap=wx.NullBitmap, label="",
